SystemOrganization addCategory: #FixedDecimal!

!Fraction class methodsFor: '*FixedDecimal' stamp: 'cbc 6/13/2005 15:22'!
readFromString: aString
	"Create an object based on the contents of aString."
	| string |

	string := (aString includes: $/) ifTrue: [aString] ifFalse: [aString, '/1'].
	^self readFrom: (ReadStream on: string)! !

!Integer methodsFor: '*FixedDecimal' stamp: 'cbc 8/30/2007 20:07'!
adaptToFixedDecimal: fixedDecimal andSend: arithmeticOpSelector
	^fixedDecimal perform: arithmeticOpSelector with: (self asFixedDecimal: 0)! !

Number subclass: #FixedDecimal
	instanceVariableNames: 'number scale part1 part2 negative'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'FixedDecimal'!

!FixedDecimal commentStamp: 'cbc 8/30/2007 20:16' prior: 0!
A FixedDecimal is similar to a ScaledDecimal, but different in certain select ways.  It's primary purpose was to be able to represent precise decimals for such things as representing money - where ScaledDecimals leave something to be desired.  For instance, with ScaledDecimals, you get:
(33.333s withScale:2) + (33.333s withScale:2)   print it yields 66.67s 
but with FixedDecimals, you would get:
(33.333 asFixedDecimal: 2) + (33.333 asFixedDecimal: 2)	print it yields 66.66.
So, FixedDecimals round the numbers to the exact scale you specify - converting a float to a FixedDecimal and back will not necessarily return the starting number, unlike ScaledDecimals.

Instance Variables
	negative:		Indicates wheither the FixedDecimal is negative.  Used for printing internally.
	number:		A raw integer reprsenting the number (not scaled!!).
	part1:		The integer part of the number.
	part2:		The non-integer part of the number (can't remember what that is called right now - too late at night).
	scale:		The scale of the FixedDecimal
!

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'cbc 8/22/2006 16:50'!
newFromFixedDecimal: aDecimal scale: scaleIn 
	"Answer a new instance of me."
	| part2 |

	part2 := (aDecimal part2 * (10 raisedToInteger: (scaleIn - aDecimal scale))) rounded.
	^self basicNew setPart1: aDecimal part1 part2: part2 scale: scaleIn negative: (aDecimal < FixedDecimal zero)! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'cbc 8/15/2006 11:42'!
newFromNumber: aNumber scale: scaleIn 
	"Answer a new instance of me."
	| temp part1 part2 |

	(aNumber isKindOf: FixedDecimal) ifTrue: [^self newFromFixedDecimal: aNumber scale: scaleIn].
	temp := aNumber asFraction.
	part1 := temp truncated.
	part2 := (temp fractionPart * (10 raisedToInteger: scaleIn)) rounded.
	part2 asString size > scaleIn ifTrue: ["Rounding can get us here..."
		temp := (temp * (10 raisedToInteger: scaleIn)) rounded / (10 raisedToInteger: scaleIn).
		part1 := temp truncated.
		part2 := (temp fractionPart * (10 raisedToInteger: scaleIn)) rounded.
		].
	^self basicNew setPart1: part1 part2: part2 scale: scaleIn negative: (temp < 0)! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'Merge version 5 and 6. 10/10/2005 15:48'!
newFromString: aNumberString scale: scaleIn
	"Answer a new instance of me.
	String is assumed to be of the format: '293492.230' "
	| temp part1 part2 |

	temp := aNumberString readStream.
	part1 := (temp upTo: $.) asSignedInteger.
	temp atEnd ifTrue: [	^self basicNew setPart1: part1 part2: 0 scale: scaleIn negative: (part1 < 0)].
	part2 := temp next: scaleIn.
	part2 := part2 asInteger * (10 raisedTo: (scaleIn - part2 size)).
	^self basicNew setPart1: part1 part2: part2 scale: scaleIn negative: (aNumberString includes: $-)! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'kph 2/2/2008 00:43'!
readFrom: aString
	"compatability with ScaledDecimal"
	
	^ self newFromString: aString scale: (aString size - (aString indexOf: $.))
	
"
FixedDecimal readFrom: '123.456' 
"! !

!FixedDecimal class methodsFor: 'testing' stamp: 'cbc 8/15/2006 11:38'!
testBoundaryIssues
	self assert: [(FixedDecimal newFromNumber: (407584 * 100 / 5112120) scale: 1) printString = '8.0']! !

!FixedDecimal class methodsFor: 'as yet unclassified' stamp: 'cbc 8/22/2006 16:48'!
zero
	^self newFromNumber: 0 scale: 0! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/30/2007 19:41'!
* operand 

	^((operand isKindOf: ScaledDecimal) and: [operand scale > scale]) ifTrue: [
		FixedDecimal newFromNumber: (operand asFraction * self asFraction) scale: operand scale
	] ifFalse: [
		FixedDecimal newFromNumber: (operand * self asFraction) scale: (scale)
	].! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/30/2007 19:42'!
+ operand 

	^((operand isKindOf: ScaledDecimal) and: [operand scale > scale]) ifTrue: [
		FixedDecimal newFromNumber: (operand asFraction + self asFraction) scale: operand scale
	] ifFalse: [
		FixedDecimal newFromNumber: (operand + self asFraction) scale: (scale)
	].! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/30/2007 19:36'!
- operand 

	^((operand isKindOf: FixedDecimal) and: [operand scale > scale]) ifTrue: [
		FixedDecimal newFromNumber: (self asFraction - operand asFraction) scale: operand scale
	] ifFalse: [
		^self + operand negated
	].! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/30/2007 19:42'!
/ operand 

	operand = 0 ifTrue: [^ (ZeroDivide dividend: self) signal].
	^((operand isKindOf: ScaledDecimal) and: [operand scale > scale]) ifTrue: [
		FixedDecimal newFromNumber: (self asFraction / operand asFraction) scale: operand scale
	] ifFalse: [
		FixedDecimal newFromNumber: (self asFraction / operand asFraction) scale: (scale)
	].! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 8/30/2007 19:40'!
// operand
	"Anser the integer quotient after dividing the receiver by operand
	with truncation towards negative infinity."
	^scale = 0 ifTrue: [part1 // operand] ifFalse: [self asFraction // operand]! !

!FixedDecimal methodsFor: 'comparing' stamp: 'cbc 2/20/2006 15:14'!
< aNumber
	"Implementation of Number 'comparing' method."
	aNumber isNumber ifFalse: [^ false].
	(aNumber isKindOf: FixedDecimal) ifTrue: [^self lessThanDecimal: aNumber].
	^aNumber adaptToFixedDecimal: self andSend: #<! !

!FixedDecimal methodsFor: 'comparing' stamp: 'cbc 3/11/2005 16:06'!
= aNumber
	"Implementation of Number 'comparing' method."
	aNumber isNumber ifFalse: [^ false].
	(aNumber isKindOf: FixedDecimal) ifTrue: [^self equalsDecimal: aNumber].
	^aNumber adaptToFixedDecimal: self andSend: #=! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:37'!
adapt: aNumber andSend: selector
	^(aNumber asFixedDecimal: scale) perform: selector with: self! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:38'!
adaptToFloat: aFloat andSend: selector
	^self adapt: aFloat andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:38'!
adaptToFraction: aFraction andSend: selector
	^self adapt: aFraction andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:38'!
adaptToInteger: anInteger andSend: selector
	^self adapt: anInteger andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:39'!
adaptToScaledDecimal: aScaledDecimal andSend: selector
	^self adapt: aScaledDecimal andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 9/6/2007 19:40'!
adaptToString: aString andSend: selector
	^self adapt: aString andSend: selector! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 8/30/2007 18:54'!
asFloat
	^(number / (10 raisedTo: scale)) asFloat! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 8/30/2007 19:18'!
asFraction
	^number / (10 raisedTo: scale)! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 8/30/2007 19:17'!
asInteger
	^part1! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 3/11/2005 16:09'!
asScale: newScale
	scale = newScale ifTrue: [^self].
	^self class newFromFixedDecimal: self scale: newScale ! !

!FixedDecimal methodsFor: 'converting' stamp: 'cbc 8/30/2007 19:17'!
asSignedInteger
	^part1! !

!FixedDecimal methodsFor: 'private' stamp: 'cbc 8/30/2007 18:46'!
equalsDecimal: decimal
	| maxScale |

	decimal scale = scale ifTrue: [^decimal number = number].
	maxScale := decimal scale max: scale.
	^(self asScale: maxScale) = (decimal asScale: maxScale)! !

!FixedDecimal methodsFor: 'truncation and round off' stamp: 'cbc 9/6/2007 18:59'!
fractionPart
	^FixedDecimal new setPart1: 0 part2: part2 scale: scale negative: negative! !

!FixedDecimal methodsFor: 'comparing' stamp: 'cbc 9/6/2007 18:53'!
hash
	^self asFraction hash! !

!FixedDecimal methodsFor: 'private' stamp: 'cbc 8/23/2006 08:52'!
lessThanDecimal: decimal
	| maxScale |

	decimal scale = scale ifTrue: [^number < decimal number].
	maxScale := decimal scale max: scale.
	^(self asScale: maxScale) < (decimal asScale: maxScale)! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 9/6/2007 19:05'!
negated
	^FixedDecimal new setPart1: part1 part2: part2 scale: scale negative: negative not! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 3/11/2005 16:12'!
number
	^number! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 3/11/2005 16:11'!
part1
	^part1! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 3/11/2005 16:11'!
part2
	^part2! !

!FixedDecimal methodsFor: 'printing' stamp: 'Merge version 5 and 6. 10/10/2005 14:58'!
printOn: aStream 
	| p2 |

	negative ifTrue: [aStream nextPut: $-].
	scale = 0 ifTrue: [aStream nextPutAll: part1 asString. ^self].
	p2 := part2 asString.
	p2 := String streamContents: [:s| 
		p2 size + 1 to: scale do: [:x| s nextPut: $0].
		s nextPutAll: p2.
		].
	aStream nextPutAll: part1 asString; nextPut: $.; nextPutAll: p2.! !

!FixedDecimal methodsFor: 'printing' stamp: 'Merge version 5 and 6. 10/10/2005 14:59'!
printOn: aStream precision: precise
	| p2 p1 |

	p1 := part1 asString.
	p1 size + scale > precise ifTrue: [^nil].
	p1 := String streamContents: [:s|
		negative ifTrue: [s nextPut: $-].
		p1 size + 1 to: (precise - scale) do: [:x| s nextPut: $0].
		precise - scale > 0 ifTrue: [s nextPutAll: p1].
		].
	scale = 0 ifTrue: [aStream nextPutAll: p1. ^self].
	p2 := part2 asString.
	p2 := String streamContents: [:s| 
		p2 size + 1 to: scale do: [:x| s nextPut: $0].
		s nextPutAll: p2.
		].
	aStream nextPutAll: p1; nextPut: $.; nextPutAll: p2.! !

!FixedDecimal methodsFor: 'printing' stamp: 'Merge version 5 and 6. 10/10/2005 14:58'!
printOn: aStream precision: precise reserveSign: boolean
	"ReserveSign, if true, means that if we are not negative, then we should preceed the number with a 0.  So that the numbers will ALWAYS be the same size when printed."
	| p2 p1 |

	p1 := part1 asString.
	p1 size + scale > precise ifTrue: [Error signal: 'Decimal is too big to print!!'].
	negative ifTrue: [aStream nextPut: $-] ifFalse: [boolean ifTrue: [aStream nextPut: $+]].
	p1 size + 1 to: (precise - scale) do: [:x| aStream nextPut: $0].
	precise - scale > 0 ifTrue: [aStream nextPutAll: p1].
	scale = 0 ifTrue: [
		boolean ifTrue: [aStream nextPut: $.].
		^self
		].
	aStream nextPut: $..
	p2 := part2 asString.
	p2 size + 1 to: scale do: [:x| aStream nextPut: $0].
	aStream nextPutAll: p2.
! !

!FixedDecimal methodsFor: 'printing' stamp: 'cbc 10/14/2005 14:40'!
printString
	^String streamContents: [:s| self printOn: s]! !

!FixedDecimal methodsFor: 'arithmetic' stamp: 'cbc 9/6/2007 18:49'!
reciprocal
	^FixedDecimal newFromNumber: (self asFraction reciprocal) scale: scale! !

!FixedDecimal methodsFor: 'accessing' stamp: 'cbc 3/11/2005 16:12'!
scale
	^scale! !

!FixedDecimal methodsFor: 'private' stamp: 'Merge version 5 and 6. 10/10/2005 15:48'!
setPart1: part1In part2: part2In scale: scaleIn negative: boolean

	part1 := part1In abs.
	part2 := part2In abs.
	scale := scaleIn.
	negative := boolean.
	number := (part1 * (10 raisedToInteger: scale)) + part2.
	negative ifTrue: [number := 0 - number].! !

!FixedDecimal methodsFor: 'truncation and round off' stamp: 'cbc 6/13/2005 15:38'!
truncated
	^part1! !

!Number methodsFor: '*FixedDecimal' stamp: 'cbc 8/30/2007 20:07'!
adaptToFixedDecimal: fixedDecimal andSend: arithmeticOpSelector
	^fixedDecimal perform: arithmeticOpSelector with: (self asFixedDecimal: fixedDecimal scale)! !

!Number methodsFor: '*FixedDecimal' stamp: 'cbc 8/30/2007 19:58'!
asFixedDecimal: scale
	^FixedDecimal newFromNumber: self scale: scale! !

!ScaledDecimal methodsFor: '*FixedDecimal' stamp: 'cbc 8/30/2007 20:07'!
adaptToFixedDecimal: fixedDecimal andSend: arithmeticOpSelector
	^fixedDecimal perform: arithmeticOpSelector with: (FixedDecimal newFromNumber: fraction scale: scale)! !

!String methodsFor: '*FixedDecimal' stamp: 'cbc 9/6/2007 19:40'!
asFixedDecimal: scale
	^FixedDecimal newFromString: self scale: scale! !
